from easypy.exceptions import TException
from easypy.timing import PredicateNotSatisfied


class CommandExecutionError(TException):
    template = "Command execution error: {name}"


class UnknownChannel(CommandExecutionError):
    template = "Unknown channel {channel}"


class UnknownExitCode(CommandExecutionError):
    template = "Unknown exit code {exit_code}"


class CommandTimeoutError(CommandExecutionError):
    template = "Command timed out after {timeout}: {name}"


class CommandPidTimeoutError(TimeoutError):
    pass


class CommandAlreadyDone(TException):
    template = "Command is already done"


class CommandAbortedByReboot(CommandExecutionError):
    template = "Command aborted by reboot request: {name}"


class CommandAbortedByOverflow(CommandExecutionError):
    template = "Command aborted because it had too much output: {name}"


class CommandOrphaned(CommandExecutionError):
    template = "Talker agent restarted while command was running: {name}"


class TalkerError(CommandExecutionError):
    template = "Talker agent encountered an error: `{_exception}`"


class CommandHanging(CommandExecutionError):
    template = "Command killed but did not die: {name}"


class ClientCommandTimeoutError(CommandExecutionError):
    template = "Command timed out after {timeout!r}: {name} - probably hanging on the host"


class TalkerClientSendTimeout(CommandExecutionError):
    template = "Talker client did not send the command within {timeout!r}: this machine is probably overloaded"


class TalkerServerTimeout(CommandExecutionError):
    template = "Talker agent did not receive command after {timeout!r}: either host or talker is down"


class TalkerCommandLost(CommandExecutionError):
    template = "Talker agent did not receive command after {timeout!r}, though host and talker are alive"


class CommandTimeout(CommandExecutionError):
    template = "Command did not finish within {timeout!r}: {name} - probably hanging on the host"


class CommandLineTimeout(CommandExecutionError):
    template = "Command did not send output for {line_timeout!r}: {name}"


class HostStillAlive(TException):
    template = "Host {hostname} did not went down within {elapsed}"


class HostDidNotRecover(CommandExecutionError):
    template = "Host {hostname} did not recover from reboot within {timeout}"


class TalkerProcessLineTimedOut(TException):
    template = "{_msg} ({machine})"


class RedisTimeoutError(TException):
    template = "Talker's redis server connection timed out"


class NoResponseForRedisCommand(TException, PredicateNotSatisfied):
    template = "No response for redis command within timeout: {cmd}"


class RedisConnectionError(TException):
    template = "Connection error while communicating with Talker Redis"



import time
import json
import logging
from uuid import uuid4
from contextlib import ExitStack

from easypy.units import DAY, HOUR, Duration
from easypy.logging import log_context
from easypy.humanize import compact
from easypy.properties import safe_property
from easypy.timing import Timer

from talker.config import (
    _logger, _verbose_logger, get_logger,
    MAX_OUTPUT_PER_CHANNEL, JOB_PID_TIMEOUT, AGENT_SEND_TIMEOUT,
    AGENT_ACK_TIMEOUT, TALKER_COMMAND_LOST_RETRY_ATTEMPTS, COMMANDS_KEY_TIMEOUT
)
from talker.errors import (
    CommandTimeout, UnknownChannel, TalkerCommandLost, HostDidNotRecover, HostStillAlive,
    TalkerClientSendTimeout, TalkerServerTimeout, ClientCommandTimeoutError, CommandTimeoutError,
    CommandAbortedByReboot, CommandAbortedByOverflow, CommandHanging, CommandOrphaned,
    CommandLineTimeout, CommandExecutionError, UnknownExitCode, TalkerError, CommandPidTimeoutError, CommandAlreadyDone
)


class AbortedBy:
    timeout = 'timeout'
    reboot = 'reboot'
    overflowed = 'overflowed'
    orphaned = 'orphaned'
    error = 'error'
    hanging = 'hanging'
    line_timeout = 'line_timeout'


class Cmd(object):
    """
    A class to represent a command to run on a host

    :param [str] host_id: Host ID to run the command on
    :param [TalkerClient] talker: The Talker client that is used to send the command
    :param [bool] raise_on_failure: Should the command raise an exception on failure
    :param [Tuple[int]] retcode: What return codes are considered success
    :param [List[str]] args: List of the command arguments
    :param [int, float] ack_timeout: Agent ack timeout in seconds
    :param [float, int] timeout: Command timeout in seconds
    :param [float, int] server_timeout: Command timeout on the agent in seconds
    :param [float, int] line_timeout: Command output timeout in seconds
    :param [str] log_file: Log file path to write command output to in the host
    :param [str] name: Command name for logging purposes
    :param [int] max_output_per_channel: Maximum size of the command output to generate in bytes
    :param [str] set_new_logpath: Set a new path for the Talker agent log file
    """

    def __init__(self,
                 host_id,
                 talker,
                 raise_on_failure=True,
                 retcode=(0,),
                 args=None,
                 ack_timeout=AGENT_ACK_TIMEOUT,
                 timeout=HOUR,
                 server_timeout=True,
                 line_timeout=None,
                 log_file=None,
                 name=None,
                 max_output_per_channel=None,
                 set_new_logpath=None,
                 ):
        self.name = name
        self.job_id = str(uuid4())
        self.talker = talker
        self.retcode = None
        self.aborted_by = None
        self.ack = None
        self.stdout = []
        self.stderr = []
        self.raise_on_failure = raise_on_failure
        self.good_codes = retcode
        self.args = args
        self.host_id = host_id
        self.hostname = host_id.partition('.')[-1]
        self.ack_timer = None
        self.handling_timer = None
        self.ack_timeout = ack_timeout
        self.timeout = timeout or HOUR  # No command should run over hour, unless specified explicitly
        self.server_timeout = server_timeout
        self.line_timeout = Duration(line_timeout) if line_timeout else None
        self.log_file = log_file
        self.max_output_per_channel = max_output_per_channel or MAX_OUTPUT_PER_CHANNEL
        self.set_new_logpath = set_new_logpath
        self.client_timer = Timer(expiration=self.timeout if not server_timeout else self.timeout + 5)
        self._line_timeout_synced = False
        self.attempts = 0  # in case of TalkerCommandLost

    def __repr__(self):
        return "%s(<%s>%s)" % (self.__class__.__name__, self.job_id, "!" if not self.ack else "")

    @property
    def _result_key(self):
        return "result-%s" % self.job_id

    @property
    def _ack_key(self):
        return "%s-ack" % self._result_key

    @property
    def _stdout_key(self):
        return "%s-stdout" % self._result_key

    @property
    def _stderr_key(self):
        return "%s-stderr" % self._result_key

    @property
    def _exit_code_key(self):
        return "%s-retcode" % self._result_key

    @safe_property
    def _pid_key(self):
        return "%s-pid" % self._result_key

    @property
    def _commands_key(self):
        return "commands-%s" % self.host_id

    @property
    def timed_out(self):
        """
        Is this command timed out

        :rtype: bool
        """
        return self.aborted_by == AbortedBy.timeout

    @property
    def since_started(self):
        """
        Get the time since the command acked

        :rtype: Duration
        """
        return Duration(time.time() - self.ack) if self.ack else None

    def get_pid(self, timeout=JOB_PID_TIMEOUT, wait_for_ack=True):
        """
        Get the command process ID

        :param [int, float] timeout: After how many seconds should we throw a timeout error
        :param [bool] wait_for_ack: Wait for ack before checking for the command PID

        :returns: The command PID
        :rtype: int
        """
        if wait_for_ack:
            self.wait(for_ack=True)

        res = self.talker.reactor.get(self._pid_key, timeout=timeout, _cmd_id=self.job_id)
        if res is None:
            if self.retcode is not None:
                raise CommandAlreadyDone()
            else:
                raise CommandPidTimeoutError('redis timeout after {} when trying to get pid'.format(timeout))

        return int(res)

    def put_command(self, *, job_repr, **job_data):
        """
        Put the command in the commands queue
        For internal use

        :param [str] job_repr: A string to print the command for logging purposes
        :param [dict] job_data: Command data to send to the Talker agent
        """
        get_logger().debug("submitting job [%s] (%s)", job_repr, self.job_id)
        self.talker.reactor.rpush(self._commands_key, json.dumps(job_data), _callback=self.on_sent, _cmd_id=self.job_id)
        # making expiration long enough so it's unlikely to expire in the middle pushing a new command - see WEKAPP-86513
        # a 5m expiration could fall exactly when we push a new command, causing it to expire before the agent pops it out;
        # a 1d expiration is long enough since new commands are likely to be pushed in the interim, extending the expiration
        # we can't call 'expire' before 'rpush', since it will be a no-op if the key doesn't exist yet
        self.talker.reactor.expire(self._commands_key, COMMANDS_KEY_TIMEOUT, _cmd_id=self.job_id)
        self.handling_timer = Timer(expiration=AGENT_SEND_TIMEOUT)

    @log_context(host="{.hostname}")
    def send(self):
        """
        Send the command to the Talker agent on the host
        """
        args = [str(a) for a in self.args]
        command_string = ' '.join(x if ' ' not in x else '"%s"' % x for x in args)
        timeout = self.timeout if self.server_timeout else None

        extras = {}
        extras.update(
            max_output_per_channel=self.max_output_per_channel,
            set_new_logpath=self.set_new_logpath,
            line_timeout=self.line_timeout,
            log_file=self.log_file,
        )

        self.put_command(id=self.job_id, cmd=args, timeout=timeout, job_repr=compact(command_string, 100), **extras)

    def on_sent(self):
        if not self.ack_timer:
            self.ack_timer = Timer(expiration=self.ack_timeout)  # this is expiration on the agent ack'ing the command

    @property
    def is_sent(self):
        """
        Is the command sent to the host

        :rtype: bool
        """
        return bool(self.ack_timer)

    @log_context(host="{.hostname}")
    def reset_server_timeout(self):
        """
        Reset timeout on the Talker agent for this command
        """
        _logger.debug("Reseting timeout, %s, %s", self.job_id, self.timeout)
        self.client_timer.reset()
        self.talker.reactor.rpush(self._commands_key, json.dumps(dict(
            id=self.job_id,
            cmd="reset_timeout",
            new_timeout=self.timeout
        )), _cmd_id=self.job_id)
        self.talker.reactor.expire(self._commands_key, COMMANDS_KEY_TIMEOUT, _cmd_id=self.job_id)

    @log_context(host="{.hostname}")
    def send_signal(self, sig):
        """
        Send a signal for the process running te command

        :param [int] sig: The signal to send to the command
        """
        self.talker.reactor.rpush(self._commands_key, json.dumps(dict(
            id=self.job_id,
            cmd="signal",
            signal=sig,
        )), _cmd_id=self.job_id)
        self.talker.reactor.expire(self._commands_key, COMMANDS_KEY_TIMEOUT, _cmd_id=self.job_id)

    @log_context(host="{.hostname}")
    def kill(self, graceful_timeout=3):
        """
        Kill the process running the command

        This will send a SIGTERM signal to the process, wait for graceful_timeout seconds,
        check if the process died and if not, send the SIGKILL signal.

        :param [int, float] graceful_timeout: Seconds to wait between the SIGTERM and SIGKILL signals
        """
        self.talker.reactor.rpush(self._commands_key, json.dumps(dict(
            id=self.job_id,
            cmd="kill",
            graceful_timeout=graceful_timeout,
        )), _cmd_id=self.job_id)
        self.talker.reactor.expire(self._commands_key, COMMANDS_KEY_TIMEOUT, _cmd_id=self.job_id)

    def _sync_timeout(self, line_timeout):
        if self._line_timeout_synced:
            return
        else:
            self._line_timeout_synced = True

        if line_timeout and line_timeout > self.timeout:
            self.timeout = int(line_timeout * 1.15)
            self.reset_server_timeout()

    def iter_results(self, line_timeout=DAY, timeout=DAY):
        """
        Iterate return code and output results

        :param [int, float] line_timeout: Timeout for new output to be recieved in seconds
        :param [int, float] timeout: Timout for command to finish in seconds

        :returns: A Tuple with the channel name and value. Channel names are `retcode`, `stdout` and `stderr`
        :rtype: Tuple[str, str]
        """
        timeout = timeout or DAY
        line_timeout = line_timeout or DAY
        blpop_timeout = min(line_timeout, timeout // 10, self.timeout // 10, AGENT_ACK_TIMEOUT // 10)

        session_timer = Timer(expiration=timeout)  # timeout for this invocation (not entire command duration)
        line_timer = Timer(expiration=line_timeout)
        timeout_reset_timer = Timer(expiration=max(line_timeout // 10, 1))
        self._sync_timeout(line_timeout=line_timeout)

        while self.retcode is None:

            if line_timer.expired:
                self.raise_exception(exception_cls=CommandLineTimeout, line_timeout=Duration(line_timeout))

            if session_timer.expired:
                self.raise_exception(exception_cls=CommandTimeout, timeout=timeout)

            channels = {
                self._ack_key: 'ack', self._stderr_key: 'stderr',
                self._stdout_key: 'stdout', self._exit_code_key: 'retcode'
            }
            res = self.talker.reactor.blpop(
                [self._ack_key, self._stderr_key, self._stdout_key, self._exit_code_key],
                timeout=blpop_timeout, _cmd_id=self.job_id
            )
            if not res:
                if not self.ack:
                    # no point checking these timers timeout if the command hasn't been acknowledged yet
                    line_timer.reset()
                    session_timer.reset()

                self.check_client_timeout()
                continue
            else:
                channel_name, value = res
                channel_name = channels[channel_name.decode('utf-8')]

            line_timer.reset()
            if timeout_reset_timer.expired:
                timeout_reset_timer.reset()
                self.reset_server_timeout()

            if channel_name == 'stderr':
                self.on_output(self.stderr, (value,))
            elif channel_name == 'stdout':
                self.on_output(self.stdout, (value,))
            elif channel_name == 'retcode':
                # see if any leftovers after receiving exit code
                stdout, stderr = self.get_output()
                for data_channel_name, values in (('stdout', stdout), ('stderr', stderr)):
                    for data_value in values:
                        yield (data_channel_name, data_value)
                value = self.set_retcode(value)
            elif channel_name == 'ack':
                self.on_ack(value)
                continue

            yield (channel_name, value)

            self.raise_if_needed()

    def iter_lines(self, line_timeout=DAY, timeout=DAY):
        """
        Iterate command output

        :param [int, float] line_timeout: Timeout for new output to be recieved in seconds
        :param [int, float] timeout: Timout for command to finish in seconds

        :returns: A Tuple containing a line from stdout or None and a line from stderr or None
        :rtype: Tuple[str, str]
        """
        done = False
        parts = {
            'stdout': '',
            'stderr': ''
        }

        def return_for_channel(channel, line):
            if channel == 'stdout':
                return (line, None)
            elif channel == 'stderr':
                return (None, line)

        def yield_channel(channel):
            lines = parts[channel].splitlines(keepends=True)
            parts[channel] = ''
            for line in lines:
                if line.endswith('\n'):
                    yield return_for_channel(channel, line.strip('\n'))
                else:
                    parts[channel] = line
            if done:
                yield return_for_channel(channel, parts[channel])

        for channel, value in self.iter_results(line_timeout=line_timeout, timeout=timeout):
            if channel in ['stdout', 'stderr']:
                try:
                    parts[channel] += value.decode('utf-8')
                except UnicodeDecodeError:
                    _logger.warning("Unable to decode %s to utf-8!", value)
                    parts[channel] += "%s" % value
                yield from yield_channel(channel)
            elif channel == 'retcode':
                done = True

    @log_context(host="{.hostname}")
    def log_pipe(self, logger, stdout_lvl=logging.INFO, stderr_lvl=logging.ERROR, line_timeout=DAY):
        """
        Pipe command output to a logger

        :param [logging.Logger] logger: The logger to use to log command output
        :param [int] stdout_lvl: Logging level to use in order to log stdout
        :param [int] stderr_lvl: Logging level to use in order to log stderr
        :param [int, float] line_timeout: Timeout for new output to be recieved in seconds
        """
        for stdout, stderr in self.iter_lines(line_timeout=line_timeout):
            if stdout_lvl is not None and stdout:
                logger.log(stdout_lvl, stdout)
            if stderr_lvl is not None and stderr:
                logger.log(stderr_lvl, stderr)

    @log_context(host="{.hostname}")
    def foreground(self, stdout=True, stderr=True):
        """
        Print command output as it is recieved from the agent

        :param [bool] stdout: Should stdout be printed
        :param [bool] stderr: Should stderr be printed

        :returns: The command return code when done
        :rtype: int
        """
        should_print = dict(stdout=stdout, stderr=stderr)
        for channel, value in self.iter_results():
            if channel in ['stdout', 'stderr'] and should_print[channel]:
                print(value.decode('utf-8'), end='')
            elif channel == 'retcode':
                return self.retcode
            else:
                self.raise_exception(exception_cls=UnknownChannel, channel=channel)

    def check_client_timeout(self):
        """
        Check if the command has timed out

        This will raise an appropriate exception depends on what has timed out:
        * If the reactor haven't sent the command before the handling timer expired: TalkerClientSendTimeout
        * If the command has yet to be acked and the agent is alive: TalkerCommandLost
        * If the command has yet to be acked and the agent is not responding: TalkerServerTimeout
        * If ack is not supported and the command has expired: ClientCommandTimeoutError
        """
        if not self.is_sent:
            # client/reactor did not send command yet
            if self.handling_timer.expired:
                self.raise_exception(exception_cls=TalkerClientSendTimeout, timeout=self.handling_timer.elapsed)
        elif not self.ack:
            is_talker_alive_name = 'is_talker_alive'

            # client did not receive command - check all timers
            if self.ack_timer.expired or self.client_timer.expired:
                talker_alive = False
                if self.name != is_talker_alive_name:  # prevent recursion
                    if self.talker.is_alive(self.host_id, is_talker_alive_name):
                        # check again, since talker may have just booted
                        self.poll(check_client_timeout=False)
                        if self.ack:
                            # seems like our command was finally answered
                            _logger.debug("we thought TalkerCommandLost, but finally not so... (%s)", self)
                            return
                        talker_alive = True
                        if self.attempts < TALKER_COMMAND_LOST_RETRY_ATTEMPTS:
                            self.attempts += 1
                            self.ack_timer = None
                            self.send()
                            return

                exception_cls = TalkerCommandLost if talker_alive else TalkerServerTimeout
                self.raise_exception(exception_cls=exception_cls, timeout=self.ack_timer.elapsed)
        elif self.client_timer.expired:
            self.raise_exception(
                exception_cls=ClientCommandTimeoutError, timeout=self.ack_timer.elapsed, started=self.since_started
            )

    def raise_if_needed(self):
        """
        Raise an error for the command failure. If the command succeeded, do nothing
        """
        if self.aborted_by == AbortedBy.timeout:
            self.raise_exception(exception_cls=CommandTimeoutError, timeout=Duration(self.timeout))
        if self.aborted_by == AbortedBy.reboot:
            self.raise_exception(exception_cls=CommandAbortedByReboot)
        if self.aborted_by == AbortedBy.overflowed:
            self.raise_exception(exception_cls=CommandAbortedByOverflow, max_output_per_channel=self.max_output_per_channel)
        if self.aborted_by == AbortedBy.hanging:
            self.raise_exception(exception_cls=CommandHanging)
        if self.aborted_by == AbortedBy.orphaned:
            self.raise_exception(exception_cls=CommandOrphaned)
        if self.aborted_by == AbortedBy.line_timeout:
            self.raise_exception(exception_cls=CommandLineTimeout, line_timeout=self.line_timeout)
        if self.aborted_by == AbortedBy.error:
            _, traceback = self.get_output(new=False)
            traceback = self._decode_output(traceback, safe=True).strip()
            exception = traceback.splitlines()[-1]
            raise TalkerError(
                host_id=self.host_id, hostname=self.hostname, fulltb=traceback, talker=self.talker, _exception=exception
            )
        if self.retcode is None:
            return
        if self.raise_on_failure and self.retcode not in self.good_codes:
            self.raise_exception()

    def raise_exception(self, exception_cls=CommandExecutionError, **params):
        """
        Raise an exception for this command
        """

        def process_output(d):
            text = self._decode_output(d, safe=True)
            if len(text) > 1010:
                text = "...%s" % text[-1000:]
            return text

        log_files = self.log_file if isinstance(self.log_file, (tuple, list)) else [self.log_file] * 2
        stds = [None] * 2 if all(log_files) else (process_output(std) for std in self.get_output(new=False))
        for (param, std, log_file) in zip("stdout stderr".split(), stds, log_files):
            params[param] = ("<%s>" % log_file) if log_file else std

        raise exception_cls(
            cmd=self.job_id,
            name=self.name or "(unnamed command)",
            args=" ".join(map(str, self.args)),
            host_id=self.host_id,
            hostname=self.hostname,
            retcode=self.retcode,
            since_started=self.since_started,
            talker=self.talker,
            **params
        )

    def on_polled(self, exit_code):
        """
        Report return code for the command 

        :param [int] exit_code: The command return code

        :returns: The command return code or None if not yet set
        :rtype: int
        """
        if self.retcode is not None:
            return self.retcode
        if exit_code is None:
            return
        _verbose_logger.debug("%s: exit-code received (%s)", self, exit_code)
        self.set_retcode(exit_code)
        return self.retcode

    def poll(self, check_client_timeout=True):
        """
        Check if the command is finished and return it's return code

        :param [bool] check_client_timeout: Also check if the command timed out

        :returns: The command return code if done
        :rtype: int
        """
        if self.retcode is not None:
            return self.retcode

        if not self.ack:
            ack = self.talker.reactor.lpop(self._ack_key, _cmd_id=self.job_id)
            if ack is None:
                if check_client_timeout:
                    self.check_client_timeout()
                return
            self.on_ack(ack)

        exit_code = self.talker.reactor.lpop(self._exit_code_key, _cmd_id=self.job_id)
        if exit_code is None:
            if check_client_timeout:
                self.check_client_timeout()
        return self.on_polled(exit_code)

    def set_retcode(self, exit_code):
        """
        Set the command return code

        :param [int] exit_code: The command return code
        """
        if exit_code is None:
            return
        try:
            exit_code = int(exit_code)
        except ValueError:
            exit_code = exit_code.decode('utf-8')

        self.on_sent()  # sometimes we get here before the callback is called
        self.ack_timer.stop()

        if isinstance(exit_code, int):
            self.retcode = exit_code
        elif hasattr(AbortedBy, exit_code):
            self.aborted_by = getattr(AbortedBy, exit_code)
            self.retcode = -1
        else:
            self.raise_exception(exception_cls=UnknownExitCode, exit_code=exit_code)
        self.raise_if_needed()

    def wait(self, for_ack=False):
        """
        Wait until the command is done

        :param [bool] for_ack: Wait for the command to be acked instead of command to finish

        :returns: The command return code when done
        :rtype: int
        """
        while True:

            if for_ack:
                if self.ack is not None:
                    return self.ack

            if self.retcode is not None:
                return self.retcode

            # we don't want to wait too long, cause we want to raise timeout exceptions promptly
            blpop_timeout = (
                1 if not self.is_sent
                else self.ack_timer.expiration // 10 if not self.ack
                else self.client_timer.expiration // 10)

            result = self.talker.reactor.blpop(
                [self._exit_code_key, self._ack_key], timeout=blpop_timeout, _cmd_id=self.job_id)
            if result is None:
                self.check_client_timeout()
                continue

            key, value = result
            key = key.decode()
            if key == self._ack_key:
                _verbose_logger.debug("%s: ack received (%s)", self, value)
                self.on_ack(value)
                continue

            return self.on_polled(result[1])

    def result(self, decode='utf8'):
        """
        Wait until the command is done and return it's stdout output

        :param [str] decode: Expected output encoding

        :returns: The command stdout output
        :rtype: str
        """
        self.wait()
        stdout, _ = self.get_output(new=False)
        return self._decode_output(stdout, decode=decode)

    def _decode_output(self, output, decode='utf8', safe=False):
        ret = b''.join(output)
        if decode:
            ret = ret.decode(decode, errors='replace' if safe else 'strict')
        return ret

    def _request_outputs(self, pipeline):
        pipeline.lrange(self._stdout_key, 0, -1)
        pipeline.lrange(self._stderr_key, 0, -1)

    def _fetch_outputs(self):
        with self.talker.reactor.pipeline() as p:
            self._request_outputs(p)
            new_stdout, new_stderr = p.execute()
        return new_stdout, new_stderr

    def _trim_outputs(self, stdout, stderr, pipeline=None):
        len_stdout = len(stdout)
        len_stderr = len(stderr)
        with ExitStack() as stack:
            if pipeline is not None:
                p = pipeline
            else:
                p = stack.enter_context(self.talker.reactor.pipeline())
            for data_len, redis_key in \
                    ((len_stdout, self._stdout_key),
                     (len_stderr, self._stderr_key)):
                if data_len:
                    p.ltrim(redis_key, data_len, -1)
            if pipeline is None:
                p.execute()

    def on_output(self, channel, data):
        """
        Update the command output

        :param [List[bytes]] channel: The channel to update. One of `self.stdout` or `self.stderr`
        :param [bytes] data: New data to add to one of the channels
        """
        if data:
            channel.extend(data)

    def on_ack(self, ack):
        """
        Update the command ack

        :param [bytes] ack: The command start time on the host
        """
        self.ack = float(ack.decode())

    def get_output(self, new=True):
        """
        Get raw command output

        :param [bool] new: Get only new output

        :returns: Stdout and stderr from the command
        :rtype: Tuple[bytes, bytes]
        """
        new_stdout, new_stderr = self._fetch_outputs()
        self.on_output(self.stdout, new_stdout)
        self.on_output(self.stderr, new_stderr)
        self._trim_outputs(stdout=new_stdout, stderr=new_stderr)
        if new:
            return new_stdout, new_stderr
        else:
            return self.stdout, self.stderr


class RebootCmd(Cmd):
    DEFAULT_FORCE = False

    """
    A command to handle host reboot

    :param [bool] force: Should reboot forcefully (might skip sync and other host cleanups)
    """

    def __init__(self, *args, force=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.force = RebootCmd.DEFAULT_FORCE if force is None else force
        self.args = ['reboot', 'force=%s' % force]

    def send(self):
        job_repr = "reboot %r, force=%s" % (self.host_id, self.force)
        self.put_command(id=self.job_id, job_repr=job_repr, cmd='reboot', force=self.force, timeout=None)

    def check_client_timeout(self):
        try:
            super().check_client_timeout()
        except ClientCommandTimeoutError as exc:
            raise HostDidNotRecover(**exc._params)

    def raise_if_needed(self):
        if self.retcode == 1:
            raise HostStillAlive(hostname=self.hostname, elapsed=self.ack_timer.elapsed)
        super().raise_if_needed()



from datetime import timedelta
import logging

from easypy.units import MINUTE, SECOND, MiB
from easypy.threadtree import ThreadContexts


AGENT_SEND_TIMEOUT = 2 * MINUTE  # the duration we allow the client for sending our command to redis
AGENT_ACK_TIMEOUT = 20 * MINUTE  # the duration we allow the server/agent to receive and start our commands
IS_ALIVE_ACK_TIMEOUT = 5
IS_ALIVE_TIMEOUT = 5
MAX_OUTPUT_PER_CHANNEL = MiB * 10

# needs to be below AGENT_SEND_TIMEOUT, so that we don't confuse a redis timeout with a reactor timeout
REDIS_SOCKET_TIMEOUT = 120
REDIS_SOCKET_CONNECT_TIMEOUT = 60
REDIS_RETRY_ON_TIMEOUT = True
REDIS_HEALTH_CHECK_INTERVAL = 30

JOB_PID_TIMEOUT = 3 * SECOND
TALKER_COMMAND_LOST_RETRY_ATTEMPTS = 1
COMMANDS_KEY_TIMEOUT = timedelta(days=1)

TALKER_CONTEXT = ThreadContexts(defaults=dict(talker_quiet=False, ack_timeout=AGENT_ACK_TIMEOUT))

_logger = logging.getLogger(__name__)
_verbose_logger = logging.getLogger(__name__ + ".verbose")


def get_logger():
    return _verbose_logger if TALKER_CONTEXT.talker_quiet else _logger


def ack_timeout(timeout):
    return TALKER_CONTEXT(ack_timeout=timeout)



#!/usr/bin/python


"""
Talker Agent (Server-Side)
==========================

* Important:
    - keep this free of dependencies (there's only a redis dependency)
    - keep this compatible with python2.6+ (no dict comprehension)

* Packaging:
    - update the 'TALKER' version in version_info
    - ./teka pack talker

* Testing:
    - See ./wepy/devops/talker.py (client-side)

"""

import fcntl
import json
import logging
import os
import os.path
import errno
import select
import signal
import subprocess
import sys
import threading
import time
import shutil
import glob
import atexit
import random
from textwrap import dedent
from contextlib import contextmanager
from logging import getLogger
from logging.handlers import RotatingFileHandler
from threading import Lock
try:
    from configparser import ConfigParser, NoSectionError
except:  # python 2.7
    from ConfigParser import ConfigParser, NoSectionError


PY3 = sys.version_info[0] == 3

# ===========================================================================================
# Define a python2/3 compatible 'reraise' function for re-raising exceptions properly
# Since the syntax is different and would not compile between versions, we need to using 'exec'

if PY3:
    def reraise(tp, value, tb=None):
        if value is None:
            value = tp()
        if value.__traceback__ is not tb:
            raise value.with_traceback(tb)
        raise value
else:
    def exec_(_code_, _globs_=None, _locs_=None):
        """Execute code in a namespace."""
        if _globs_ is None:
            frame = sys._getframe(1)
            _globs_ = frame.f_globals
            if _locs_ is None:
                _locs_ = frame.f_locals
            del frame
        elif _locs_ is None:
            _locs_ = _globs_
        exec("""exec _code_ in _globs_, _locs_""")

    exec_(dedent("""
        def reraise(tp, value, tb=None):
            raise tp, value, tb
        """))

# ===========================================================================================

CONFIG_FILENAME = '/root/talker/config.ini'
REBOOT_FILENAME = '/root/talker/reboot.id'
EXCEPTION_FILENAME = '/root/talker/last_exception'
VERSION_FILENAME = '/root/talker/version'
JOBS_DIR = '/root/talker/jobs'
JOBS_SEEN = os.path.join(JOBS_DIR, 'eos.json')

logger = getLogger()

CYCLE_DURATION_MS = 100.0
CYCLE_DURATION = CYCLE_DURATION_MS / 1000.0
READ_ONLY = select.POLLIN | select.POLLPRI | select.POLLHUP | select.POLLERR

SIMULATE_DROPS_RATE = 0.0  # set a float (0.0 - 1.0), that will cause the server to simulate TalkerCommandLost

FINAL_RESULT_EXPIRATION = 3600

# let command output expire after some time.
# we assume that any output a command write is read shortly after, and if it isn't - it is not needed, except for troubleshooting
CMD_OUTPUT_EXPIRATION = 3600 * 24

MAX_OUTPUT_PER_CHANNEL = 10 * 2 ** 20
DEFAULT_GRACEFUL_TIMEOUT = 3

JOBS_EXPIRATION = 15  # 20 * 60  # how long to keep job ids in the EOS registry (exactly-once-semantics)

config = None
first_exception_info = None
safe_thread_lock = Lock()


class SafeThread(threading.Thread):
    def __init__(self, target, name, args=(), kwargs=None, daemon=None):
        super(SafeThread, self).__init__(None, target, name, args, kwargs)
        self.setDaemon(daemon)

    def run(self):
        global first_exception_info
        try:
            if PY3:
                self._target(*self._args, **self._kwargs)
            else:
                self._Thread__target(*self._Thread__args, **self._Thread__kwargs)
        except:
            exc_info = sys.exc_info()
            logger.info("exception in '%s'", self.name, exc_info=exc_info)
            with safe_thread_lock:
                if not first_exception_info:
                    first_exception_info = sys.exc_info()
        finally:
            # Avoid a refcycle if the thread is running a function with
            # an argument that has a member that points to the thread.
            if PY3:
                del self._target, self._args, self._kwargs
            else:
                del self._Thread__target, self._Thread__args, self._Thread__kwargs


class LineTimeout(Exception):
    pass


class JobHanging(Exception):
    pass


class JobTimeout(Exception):
    pass


class Config:
    def __init__(self, filename=CONFIG_FILENAME):
        self._filename = filename
        self._parser = ConfigParser()
        self.load_configuration()

    @property
    def parser(self):
        return self._parser

    def load_configuration(self):
        try:
            self._parser.read(self._filename)
        except (IOError, OSError) as e:
            logger.error(e, exc_info=True)
            sys.exit(1)

        logger.info("Loaded configuration: %s", self._filename)

        self.validate_config()

    def update(self, section, **kwargs):
        for key, val in kwargs.items():
            self._parser.set(section, key, val)

        with open(self._filename, 'w') as configfile:
            self._parser.write(configfile)

    def validate_config(self):
        if not self._parser.get('agent', 'host_id'):
            raise Exception('config agent.host_id is missing')


class Job(object):

    class OutputChannel(object):
        __slots__ = "name", "chunks", "closed", "total", "expiration", "fdno", "job"

        def __init__(self, name, job):
            self.name = name
            self.chunks = []
            self.closed = False
            self.total = 0
            self.expiration = 0  # expiration of data in redis
            self.fdno = None
            self.job = job

        def close(self):
            self.closed = True

        def write(self, data):
            byte_count = len(data)
            self.job.logger.debug("Got data from %s (fd=%s), len=%s", self.name, self.fdno, byte_count)

            self.total += byte_count
            self.job.last_read = time.time()
            if byte_count == 0:
                self.close()
                return True

            return self._write(data)

        def _write(self, data):
            if self.total > self.job.max_output_per_channel:
                self.job.set_result("overflowed")
                self.job.finalize()
                self.job.kill()
                return True

            self.chunks.append(data)
            return False

    class LoggedOutputChannel(OutputChannel):
        __slots__ = "log"

        def __init__(self, *args, **kwargs):
            log_file = kwargs.pop('log_file')
            super(Job.LoggedOutputChannel, self).__init__(*args, **kwargs)

            dirname = os.path.dirname(log_file)
            if dirname:
                try:
                    os.makedirs(dirname)
                except OSError as e:
                    if e.errno != errno.EEXIST or not os.path.isdir(dirname):
                        raise

            self.log = open(log_file, "ab")
            self.job.logger.info("Logging %s to '%s' (%s)", self.job.job_id, log_file, self.name)

        def close(self):
            super(Job.LoggedOutputChannel, self).close()
            self.log.close()

        def _write(self, data):
            self.log.write(data)
            return False

        def __del__(self):
            self.log.close()

    def create_channel(self, name, log_file):
        params = dict(name=name, job=self)
        cls = self.OutputChannel

        if log_file:
            params.update(log_file=log_file)
            cls = self.LoggedOutputChannel

        return cls(**params)

    def __init__(
            self, id, cmd, agent, timeout=None, buffer_time=100,
            log_file=None,  # a path or paths (stdout, stderr) for sending output to
            line_timeout=None,  # timeout for output (in either stdout or stderr) to be read from process
            set_new_logpath=None,  # used to tell talker to move its own logfile to a new location
            max_output_per_channel=MAX_OUTPUT_PER_CHANNEL,  # don't allow more than this amount of data to be sent back
            **kw):

        self.job_id = id
        self.logger = getLogger(self.job_id)

        self.logger.debug("Got job, cmd %s", cmd)
        if kw:
            # we want to allow new talker-clients to talk to older talker-agents, so we
            # politely swallow any unknown parameters that the client might have sent us
            self.logger.warning("got unsupported params: %s", tuple(map(str, sorted(kw.keys()))))

        self.cmd = cmd
        self.popen = None
        self.job_fn = None
        self.buffer_time = buffer_time  # buffer time in ms, will poll at most with this rate
        self.timeout = timeout
        self.line_timeout = line_timeout
        self.last_read = None

        out_log, err_log = log_file if isinstance(log_file, (tuple, list)) else ([log_file] * 2)
        self.stdout = self.create_channel(name="stdout", log_file=out_log)
        self.stderr = self.create_channel(name="stderr", log_file=err_log)
        self.channels = (self.stdout, self.stderr)

        self.exit_code = None
        self.killed_by = False
        self.last_send = 0
        self.agent = agent
        self.start_time = time.time()
        self.reset_timeout()
        self.max_output_per_channel = max_output_per_channel
        self.set_new_logpath = set_new_logpath

        self.acknowledge()

    def acknowledge(self):
        pipeline = self.agent.redis.pipeline()
        pipeline.rpush("result-%s-ack" % self.job_id, self.start_time)
        pipeline.expire("result-%s-ack" % self.job_id, FINAL_RESULT_EXPIRATION)
        pipeline.execute()

    def start(self):
        already_seen = self.agent.check_if_seen(self.job_id)
        if already_seen:
            self.logger.debug("already seen at %s", time.ctime(already_seen))
            return

        self.logger.debug("starting job")
        attempts = 0

        if self.agent.pending_exception:
            self.stderr.chunks.append(self.agent.pending_exception)
            self.set_result("error")
            self.finalize()
            return

        while True:
            attempts += 1
            try:
                self.popen = subprocess.Popen([u"%s" % arg for arg in self.cmd], stdout=subprocess.PIPE,
                                              stderr=subprocess.PIPE, preexec_fn=os.setsid, close_fds=True)
            except OSError as e:
                self.logger.error("OS Error:%s", e, exc_info=True)

                if attempts < 3:
                    if e.args[0] == errno.EAGAIN:
                        time.sleep(random.random() * 2)
                        continue

                self.stderr.chunks.append(e.strerror)
                self.set_result(e.errno)
                self.finalize()
                return
            else:
                break

        self.job_fn = "%s/job.%s.%s" % (JOBS_DIR, self.job_id, self.popen.pid)
        with open(self.job_fn, "w") as f:
            try:
                f.write(repr(self.cmd))
            except IOError:
                # to help with WEKAPP-74054
                os.system("df")
                os.system("df -i")
                raise

        self.agent.current_processes[self.job_id] = self
        for channel in self.channels:
            pipe = getattr(self.popen, channel.name)
            channel.fdno = fileno = pipe.fileno()
            fcntl.fcntl(fileno, fcntl.F_SETFL, os.O_NONBLOCK)
            self.agent.fds_to_channels[fileno] = channel
            self.agent.fds_to_files[fileno] = pipe
            self.agent.fds_poller.register(fileno)

        self.last_read = time.time()
        self.logger.debug("job started: pid=%s", self.popen.pid)
        pipeline = self.agent.redis.pipeline()
        pipeline.set("result-%s-pid" % self.job_id, self.popen.pid)
        pipeline.execute()

    def finalize(self, pipeline=None):
        own_pipeline = pipeline is None
        if pipeline is None:
            pipeline = self.agent.redis.pipeline()
        self.close()
        self.send_data(throttled=False, pipeline=pipeline)
        self.send_result(pipeline)
        if own_pipeline:
            pipeline.execute()

    def reset_timeout(self, new_timeout=None):
        if new_timeout:
            self.timeout = new_timeout
        self.logger.debug("Reseting timeout to %s", self.timeout)
        if new_timeout == -1:
            self.timeout = 0
            self.timeout_at = None
            return
        if self.timeout:
            self.timeout_at = time.time() + self.timeout
        else:
            self.timeout_at = None

    def check_timed_out(self):
        now = time.time()
        if not self.line_timeout or not self.last_read:
            pass
        elif now - self.last_read > self.line_timeout:
            raise LineTimeout()

        if not self.timeout_at:
            pass
        elif self.timeout_at > now:
            pass
        elif self.killed_by:
            raise JobHanging()
        else:
            raise JobTimeout()

    def send_data(self, pipeline, throttled=True):
        now = time.time()

        if throttled:
            if now - self.last_send < (self.buffer_time / 1000):
                return
            else:
                self.last_send = now

        for channel in self.channels:
            if isinstance(channel, self.LoggedOutputChannel):
                continue
            chunks = []
            while True:
                try:
                    chunks.append(channel.chunks.pop(0))
                except IndexError:
                    break
            if chunks:
                to_send = b"".join(chunks)
                self.logger.debug("Sending data from %s, len=%s", channel.name, len(to_send))
                cmd = 'result-%s-%s' % (self.job_id, channel.name)
                pipeline.rpush(cmd, to_send)
                if (channel.expiration - now) < 3600:
                    pipeline.expire(cmd, CMD_OUTPUT_EXPIRATION)
                    pipeline.expire("result-%s-pid" % self.job_id, CMD_OUTPUT_EXPIRATION)
                    channel.expiration = now + CMD_OUTPUT_EXPIRATION

    def send_result(self, pipeline):
        self.logger.debug("Job done: %s", self.exit_code)
        pipeline.rpush("result-%s-retcode" % self.job_id, self.exit_code)
        pipeline.expire("result-%s-retcode" % self.job_id, FINAL_RESULT_EXPIRATION)
        pipeline.expire("result-%s-stderr" % self.job_id, FINAL_RESULT_EXPIRATION)
        pipeline.expire("result-%s-stdout" % self.job_id, FINAL_RESULT_EXPIRATION)
        pipeline.expire("result-%s-pid" % self.job_id, FINAL_RESULT_EXPIRATION)
        if self.set_new_logpath:
            self.logger.info("Got request to set a new logpath: %s", self.set_new_logpath)
            if self.exit_code == 0:
                set_logging_to_file(self.set_new_logpath)
            else:
                self.logger.info("Skipping due to non-zero exit code")

    def set_result(self, result):
        self.logger.debug("Setting result, %s", result)
        if self.exit_code is None:
            self.exit_code = result

    def close(self):
        if self.popen:
            for fileno in (self.popen.stdout.fileno(), self.popen.stderr.fileno()):
                channel = self.agent.fds_to_channels.pop(fileno, None)
                if channel:
                    channel.close()
            self.close_popen()
        if self.job_fn:
            os.remove(self.job_fn)
        self.agent.current_processes.pop(self.job_id, None)

    def close_popen(self):
        if self.popen.stdout:
            self.popen.stdout.close()
        if self.popen.stderr:
            self.popen.stderr.close()
        if self.popen.stdin:
            self.popen.stdin.close()

    def sync_progress(self, pipeline, now):
        if all(c.closed for c in self.channels) and self.popen.poll() is not None:
            self.set_result(self.popen.poll())
            self.finalize(pipeline)

        elif self.exit_code is None:
            try:
                self.check_timed_out()
            except JobHanging:
                self.logger.error("Job did not die")
                self.set_result('hanging')
                self.finalize(pipeline)
            except JobTimeout:
                self.logger.error("Job timed out")
                self.set_result('timeout')
                self.finalize(pipeline)
                self.kill()
            except LineTimeout:
                self.logger.info("No output for too long (%s)", self.line_timeout)
                self.set_result('line_timeout')
                self.finalize(pipeline)
                self.kill()
            else:
                self.send_data(pipeline)
        else:
            self.send_data(pipeline)

    def send_signal(self, sig):
        self.logger.debug("Sending signal %s", sig)
        try:
            os.killpg(os.getpgid(self.popen.pid), sig)
        except OSError as e:
            if e.errno != errno.ESRCH:
                reraise(*sys.exc_info())

    def kill(self, graceful_timeout=DEFAULT_GRACEFUL_TIMEOUT):
        self.logger.debug("Killing job (%s)", self.popen.pid)

        def _kill():
            try:
                self.killed_by = signal.SIGTERM
                self.send_signal(signal.SIGTERM)
                time.sleep(graceful_timeout)
                if self.popen.poll() is None:
                    self.killed_by = signal.SIGKILL
                    self.send_signal(signal.SIGKILL)
            except Exception as e:
                self.logger.error(e)

        SafeThread(target=_kill, name="killer-%s" % self.job_id, daemon=True).start()
        self.reset_timeout(new_timeout=graceful_timeout + 10)


class RebootJob(Job):
    def __init__(self, *args, **kwargs):
        self.force = kwargs.pop('force')
        super(RebootJob, self).__init__(*args, **kwargs)

    def start(self):
        SafeThread(target=self.reboot_host, name="Reboot").start()

    def reboot_host(self):
        with open(REBOOT_FILENAME, 'w') as f:
            f.write(self.job_id)
        self.agent.current_processes[self.job_id] = self  # So results will be send by sender thread
        self.agent.stop_for_reboot(requested_by=self)

        proc = subprocess.Popen(['sync'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if wait_proc(proc, 60):
            self.log("Sync finished successfully")
        else:
            self.log("Sync timed out")

        self.log("Starting reboot")
        if self.force:
            self.log("Reboot using sysrq")
            proc = subprocess.Popen(
                ['bash', '-ce', 'echo 1 > /proc/sys/kernel/sysrq; echo b > /proc/sysrq-trigger'],
                stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
        else:
            self.log("Reboot with reboot cmd")
            proc = subprocess.Popen(['reboot'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)

        flush_logger()
        wait_proc(proc, timeout=60)
        time.sleep(60)

        self.log("Talker still alive, raising HostStillAlive(1)")
        time.sleep(1)  # Giving chance to report message
        self.set_result(1)
        self.finalize()
        raise Exception("Reboot did not occur")

    def log(self, s, *args):
        self.logger.info(s, *args)
        if args:
            s %= args
        self.stdout.chunks.append((s + "\n").encode('utf-8'))


class TalkerAgent(object):

    def __init__(self):
        self.output_lock = threading.RLock()
        self.redis = None
        self.host_id = None
        self.job_poller = None
        self.fds_poller = select.poll()
        self.fds_to_channels = {}
        self.fds_to_files = {}
        self.seen_jobs = {}
        self.stop_fetching = threading.Event()
        self.fetching_stopped = threading.Event()
        self.stop_agent = threading.Event()
        self.current_processes = {}
        self.exc_info = None
        self.last_scrubbed = 0

    def check_if_seen(self, job_id):
        now = time.time()
        first_seen = self.seen_jobs.setdefault(job_id, now)
        if first_seen == now:
            return
        return first_seen

    def kill(self):
        self.stop_fetching.set()
        self.stop_agent.set()

    def fetch_new_jobs(self):
        """
        format of job
            dict(
                id=str(uuid4()),
                cmd=["/bin/bash", '-c', "for i in $(seq 2);do sleep 1; echo $i; done"],
            )
        """
        last_reported = 0
        jobs_key = 'commands-%s' % self.host_id
        while not self.stop_fetching.is_set():
            new_jobs = []
            ret = self.redis.blpop([jobs_key], timeout=1)
            if not ret:
                now = time.time()
                self.scrub_seen_jobs(now=now)
                if now - last_reported > 10:
                    logger.debug("No new jobs...")
                    last_reported = now
                continue
            _, job_data_raw = ret
            new_jobs.append(job_data_raw)

            # Could be that multiple jobs were sent, checking with lrange
            additional_jobs = self.redis.lrange(jobs_key, 0, -1)
            if len(additional_jobs):
                self.redis.ltrim(jobs_key, len(additional_jobs), -1)

            new_jobs.extend(additional_jobs)
            logger.debug("Got %s jobs", len(new_jobs))
            for job in new_jobs:
                self.start_job(job.decode("utf-8"))

        self.fetching_stopped.set()
        logger.info("stopped fetching new jobs")

    def scrub_seen_jobs(self, now):
        if now - self.last_scrubbed < 10:
            return
        # for back-compat with python2.6, we don't use dict-comprehension or tuple unpacking
        self.seen_jobs = dict((job_id, last_seen) for job_id, last_seen in self.seen_jobs.items() if (now - last_seen) < JOBS_EXPIRATION)
        logger.debug("Dumping seen-jobs registry (%s items)", len(self.seen_jobs))
        with open(JOBS_SEEN, "w") as f:
            json.dump(self.seen_jobs, f)
        self.last_scrubbed = now

    def reset_timeout(self, job_id, new_timeout=None, delayed=False):
        job = self.current_processes.get(job_id)
        if job is None:
            if not delayed:
                t = threading.Timer(5, self.reset_timeout, kwargs=dict(
                    job_id=job_id,
                    new_timeout=new_timeout,
                    delayed=True))
                t.start()
                logger.error("Received timeout reset request, but job_id %s was not found, delaying for 5 seconds", job_id)
            else:
                logger.error("Delayed by 5 seconds, but job_id %s still not found", job_id)
            return
        job.reset_timeout(new_timeout=new_timeout)

    def signal_job(self, job_id, signal):
        job = self.current_processes.get(job_id)
        if job is None:
            return
        job.send_signal(signal)

    def kill_job(self, job_id, graceful_timeout):
        job = self.current_processes.get(job_id)
        if job is None:
            return
        job.kill(graceful_timeout=graceful_timeout)

    def stop_for_reboot(self, requested_by):
        self.stop_fetching.set()
        requested_by.log("Waiting for polling to stop")
        self.fetching_stopped.wait(120)
        assert self.fetching_stopped.is_set(), "Polling did not stop"
        requested_by.log("Polling stopped")
        wait_start = time.time()

        while True:
            pending = set(self.current_processes).difference((requested_by.job_id,))  # for python2 compatibility
            if not pending:
                requested_by.log("All jobs are done")
                return

            requested_by.log("Waiting for %s jobs to finish", len(pending))
            for job_id in pending:
                logger.debug("- %s", job_id)

            time.sleep(1)

            if time.time() - wait_start > 60:
                break

        requested_by.log("Some jobs not yet finished, setting exit code to 'reboot' and proceeding")
        with self.pipeline() as pipeline:
            for job_id, job in list(self.current_processes.items()):
                if job_id == requested_by.job_id:
                    continue
                job.set_result('reboot')
                job.finalize(pipeline)

    def finalize_previous_session(self):
        self.pending_exception = None
        if os.path.exists(EXCEPTION_FILENAME):
            logger.info("Found exception from previous run")
            with open(EXCEPTION_FILENAME, 'rb') as f:
                self.pending_exception = f.read()
            os.remove(EXCEPTION_FILENAME)

        with self.pipeline() as pipeline:
            for fn in glob.glob("%s/job.*.*" % JOBS_DIR):
                _, job_id, pid = fn.split(".")
                job = Job(id=job_id, cmd=None, agent=self)
                job.logger.info("Found orphaned job: %s, pid=%s", job_id, pid)
                if self.pending_exception:
                    job.stderr.chunks.append(self.pending_exception)
                    job.send_data(pipeline=pipeline, throttled=False)
                job.set_result('orphaned')
                job.finalize(pipeline=pipeline)
                os.renames(fn, "%s/orphaned.%s.%s" % (JOBS_DIR, job_id, pid))

        if os.path.exists(REBOOT_FILENAME):
            with open(REBOOT_FILENAME, 'r') as f:
                job_id = f.read()
                job = Job(id=job_id, cmd=None, agent=self)
                job.logger.info("Recovered from reboot %s", job_id)
                job.set_result(0)
                job.finalize()
            os.remove(REBOOT_FILENAME)

    def start_job(self, job_data_raw):
        job_data = json.loads(job_data_raw)
        cmd = job_data['cmd']
        if isinstance(cmd, list):
            if SIMULATE_DROPS_RATE and cmd != ['true'] and random.random() > SIMULATE_DROPS_RATE:
                logger.warning("dropping job: %(id)s", job_data)
                return
            job = Job(agent=self, **job_data)
            job.start()
        elif cmd == 'reset_timeout':
            new_timeout = job_data.get('new_timeout')
            self.reset_timeout(job_data['id'], new_timeout=new_timeout)
        elif cmd == 'reset_error':
            logger.info("Got reset error request, clearing %s", "error" if self.pending_exception else "nothing")
            self.pending_exception = None
        elif cmd == 'signal':
            sig = job_data.get('signal', signal.SIGTERM)
            self.signal_job(job_data['id'], signal=sig)
        elif cmd == 'kill':
            graceful_timeout = job_data.get('graceful_timeout', DEFAULT_GRACEFUL_TIMEOUT)
            self.kill_job(job_data['id'], graceful_timeout=graceful_timeout)
        elif cmd == 'reboot':
            logger.info("Got reboot request, force=%s", job_data.get("force"))
            job = RebootJob(agent=self, **job_data)
            job.start()
        else:
            logger.error("Cmd %s not supported, skipping", cmd)

    def get_jobs_outputs(self):
        if not self.fds_to_channels:
            return False
        to_read = self.fds_poller.poll(CYCLE_DURATION_MS)
        read_fds = set()
        for fdno, _ in to_read:
            read_fds.add(fdno)
            while True:
                file = self.fds_to_files.get(fdno)
                is_done = False

                if not file:  # closed by timeout
                    break

                try:
                    data = file.read()
                    if data is None:
                        raise IOError()  # Python3 has None on empty read, python2 raises IOError
                except IOError:
                    # no data to read
                    break
                except ValueError:
                    data = b''  # leads to cleanly closing the channel

                is_done |= len(data) == 0

                channel = self.fds_to_channels.get(fdno)  # ensuring that not closed by timeout

                # sending EOL to write func for it to handle close and breaking only then
                if channel:
                    is_done |= channel.write(data)

                if is_done:
                    self.unregister_fileno(fdno)
                    break

        return True

    @contextmanager
    def pipeline(self):
        pipeline = self.redis.pipeline()
        yield pipeline
        if len(pipeline):
            logger.debug("Sending %s commands via pipeline", len(pipeline))
            pipeline.execute()
            logger.debug("Done sending")

    def sync_jobs_progress(self):
        pipeline = self.redis.pipeline()
        now = time.time()
        while True:
            for job in list(self.current_processes.values()):
                job.sync_progress(pipeline, now=now)
            if len(pipeline):
                logger.debug("Sending %s commands via pipeline", len(pipeline))
                pipeline.execute()
                logger.debug("Done sending")
                pipeline.reset()
            else:
                time.sleep(CYCLE_DURATION)

    def start(self):
        global first_exception_info
        first_exception_info = None

        self.finalize_previous_session()
        if os.path.isfile(JOBS_SEEN):
            with open(JOBS_SEEN, "r") as f:
                self.seen_jobs = json.load(f)

        SafeThread(target=self.fetch_new_jobs, name="RedisFetcher", daemon=True).start()
        SafeThread(target=self.sync_jobs_progress, name="JobProgress", daemon=True).start()

        while not self.stop_agent.is_set():
            if not self.get_jobs_outputs():
                time.sleep(CYCLE_DURATION / 10.0)
            if first_exception_info:
                logger.debug("re-raising exception from worker")
                reraise(*first_exception_info)
                assert False, "exception should have been raised"

    def setup(self):
        host_id = config.parser.get('agent', 'host_id')
        host = config.parser.get('redis', 'host')
        port = config.parser.getint('redis', 'port')
        try:
            password = config.parser.get('redis', 'password')
        except:
            password = None

        socket_timeout = config.parser.getfloat('redis', 'socket_timeout')
        socket_connect_timeout = config.parser.getfloat('redis', 'socket_connect_timeout')
        retry_on_timeout = config.parser.getboolean('redis', 'retry_on_timeout')
        health_check_interval = config.parser.getfloat('redis', 'health_check_interval')

        logger.info("Connecting to redis %s:%s", host, port)
        import redis  # deferring so that importing talker (for ut) doesn't immediately fail if package not available
        self.redis = redis.StrictRedis(
            host=host, port=port, db=0, password=password,
            socket_timeout=socket_timeout, socket_connect_timeout=socket_connect_timeout,
            retry_on_timeout=retry_on_timeout, health_check_interval=health_check_interval
        )

        while True:
            try:
                self.redis.ping()
            except redis.ConnectionError as exc:
                logger.info("Redis not ready (%s)", exc)
                time.sleep(10)
            else:
                break

        logger.info("Connected to redis as %s", host_id)
        self.host_id = host_id

        if not os.path.isdir(JOBS_DIR):
            os.makedirs(JOBS_DIR)

    def unregister_fileno(self, fileno):
        try:
            self.fds_poller.unregister(fileno)
        except KeyError:
            pass  # closed while processing data


def run_diagnostics():
    try:
        script_path = config.parser.get('diagnostics', 'script_path')
    except NoSectionError:
        return

    if os.path.isfile(script_path):
        logger.info("Running diagnostics")
        try:
            subprocess.Popen(["bash", "-ce", script_path], close_fds=True)
        except:
            logger.exception("Running diagnostics failed")
    else:
        logger.warning("Diagnostics script file does not exist")


def wait_proc(proc, timeout):
    t = SafeThread(target=proc.wait, name='wait_proc')
    t.start()
    t.join(timeout)
    return not t.is_alive()


def handle_exception(*exc_info):
    if issubclass(exc_info[0], KeyboardInterrupt):
        sys.__excepthook__(*exc_info)
        return
    logger.error("Uncaught exception", exc_info=exc_info)


FILE_LOG_HANDLER = None


def set_logging_to_file(logpath):
    global FILE_LOG_HANDLER
    if FILE_LOG_HANDLER:
        prev_logpath = FILE_LOG_HANDLER.stream.name
        logger.info("closing %s; switched to %s", prev_logpath, logpath)
        logging.root.removeHandler(FILE_LOG_HANDLER)
        FILE_LOG_HANDLER.close()
        shutil.move(prev_logpath, logpath)

    dirname = os.path.dirname(logpath)
    for i in range(60):
        if os.path.isdir(dirname):
            break
        logger.info("%s not found, waiting to be mounted...", dirname)
        time.sleep(2)
    else:
        raise Exception("logpath directory not found: %s" % dirname)

    handler = RotatingFileHandler(logpath, encoding='utf-8', maxBytes=50 * 1024 * 1024, backupCount=2)
    handler.setFormatter(logging.Formatter("%(asctime)s|%(threadName)-15s|%(name)-38s|%(levelname)-5s|%(funcName)-15s|%(message)s"))
    logging.root.addHandler(handler)

    FILE_LOG_HANDLER = handler

    config.update('logging', logpath=logpath)


def flush_logger():
    if not FILE_LOG_HANDLER:
        return

    FILE_LOG_HANDLER.flush()
    logfile_fd = FILE_LOG_HANDLER.stream.fileno()
    os.fsync(logfile_fd)


def setup_logging(verbose):
    sys.excepthook = handle_exception
    for handler in logging.root.handlers:
        logging.root.removeHandler(handler)

    logging.root.setLevel(logging.DEBUG)

    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter("%(asctime)s|%(levelname)-5s|%(funcName)-15s|%(message)s", datefmt="%H:%M:%S"))
    if not verbose:
        handler.setLevel(logging.INFO)
    logging.root.addHandler(handler)


def main(*args):
    global config

    setup_logging(verbose="-v" in args)
    no_restart = "--no-restart" in args

    version = open(VERSION_FILENAME).read()
    logger.info("Starting Talker: %s", version)

    config = Config()
    set_logging_to_file(config.parser.get('logging', 'logpath'))

    # to help with WEKAPP-74054
    os.system("df")
    os.system("df -i")

    open("/var/run/talker.pid", "w").write(str(os.getpid()))
    atexit.register(os.unlink, "/var/run/talker.pid")

    try:
        run_diagnostics()
        agent = TalkerAgent()
        agent.setup()
        agent.start()
    except SystemExit:
        raise
    except KeyboardInterrupt:
        logger.info("(user interrupted)")
        return 2
    except Exception as exc:
        from traceback import format_exception
        try:
            with open(EXCEPTION_FILENAME, "w") as f:
                f.writelines(format_exception(*sys.exc_info()))
        except:  # noqa
            pass

        run_diagnostics()

        if not no_restart:
            logger.error("Uncaught exception - committing suicide, and restarting in 3 seconds")
            subprocess.Popen(["bash", "-ce", 'sleep 3 && service talker restart'], close_fds=True)  # make sure this matches the pattern in 'install_talker()'
        raise
    else:
        if not no_restart:
            logger.info("Restarting in 1 seconds")
            subprocess.Popen(["bash", "-ce", 'sleep 1 && service talker restart'], close_fds=True)  # make sure this matches the pattern in 'install_talker()'


if __name__ == '__main__':
    args = sys.argv[1:]
    if "--ut" in args:
        print("Talker don't need no UT")
    else:
        sys.exit(main(*args))



from collections import namedtuple
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from functools import partial
from queue import Queue, Empty
from threading import RLock, Event, Semaphore

import redis.exceptions

from easypy.concurrency import _check_exiting, concurrent, _run_with_exception_logging, raise_in_main_thread
from easypy.timing import wait, Timer
from easypy.units import MINUTE

from talker.errors import NoResponseForRedisCommand, RedisConnectionError
from talker.config import _logger, _verbose_logger, REDIS_SOCKET_TIMEOUT


class TalkerReactor():
    ASYNC_COMMANDS = {'rpush', 'expire'}
    BLOCKING_COMMANDS = {'blpop', 'brpop', 'get'}

    CmdItem = namedtuple("CmdItem", "cmd_idx cmd_id cmd args kwargs callback event results")

    def __init__(self, talker):
        self.talker = talker
        self._max_workers = 5
        self._executors = ThreadPoolExecutor(max_workers=self._max_workers)
        self._commands_queue = Queue()
        self._commands = dict()
        self._lock = RLock()
        self._current_workers = Semaphore(self._max_workers)
        self._cmd_idx = 0

        reactor_loop = _logger.context(host="TLKR-reactor-loop")(self._get_main_loop)
        self._main_loop = concurrent(func=reactor_loop, threadname="TLKR-reactor")
        self._main_loop.start()

    @staticmethod
    def _log_sent_items(items):
        for item in items:
            if item.callback:
                item.callback()

            log_message = 'reactor sent command {}'.format(item.cmd)
            if item.cmd_id:
                log_message = '{}: {}'.format(log_message, item.cmd_id)

            _verbose_logger.debug(log_message)

    def _get_main_loop(self):
        while True:
            items = []
            items.append(self._commands_queue.get())
            t = Timer(expiration=10.0 / 1000.0)
            while not t.expired:
                try:
                    items.append(self._commands_queue.get(timeout=t.remain))
                except Empty:
                    break
            if not items:
                continue
            self._current_workers.acquire(timeout=10)
            self._executors.submit(_run_with_exception_logging, self._send_data, (items,), {}, {})

    @raise_in_main_thread()
    def _send_data(self, items):
        try:
            with self.talker.redis.pipeline() as pipeline:
                for item in items:
                    redis_func = getattr(pipeline, item.cmd)
                    redis_func(*item.args, **item.kwargs)

                with RedisConnectionError.on_exception(
                        redis.exceptions.ConnectionError, redis=self.talker.redis, commands=pipeline.command_stack):
                    results = pipeline.execute()

                assert len(results) == len(items), "Our redis pipeline got out of sync?"
                self._log_sent_items(items)  # logs to debug WEKAPP-35305
                for item, result in zip(items, results):
                    if item.event:  # non-async
                        item.results.append(result)
                        item.event.set()
        finally:
            self._current_workers.release()

    def send(self, cmd, *args, _async=False, _callback=None, _cmd_id=None, **kwargs):
        with self._lock:
            self._cmd_idx += 1
            cmd_idx = self._cmd_idx

        event = None if cmd in self.ASYNC_COMMANDS else Event()
        item = self.CmdItem(
            cmd_idx=cmd_idx, cmd_id=_cmd_id, cmd=cmd,
            args=args, kwargs=kwargs, callback=_callback,
            event=event, results=[])

        if cmd not in self.ASYNC_COMMANDS:
            self._commands[cmd_idx] = item

        self._commands_queue.put(item)

        if _async:
            return cmd_idx
        if event is not None:
            return self.get_response(cmd_idx)

    def get_response(self, cmd_idx):
        item = self._commands.pop(cmd_idx)

        def has_response():
            _check_exiting()
            if not item.event.wait(timeout=0.5):
                raise NoResponseForRedisCommand(**item._asdict())
            return True

        wait(REDIS_SOCKET_TIMEOUT + MINUTE, has_response, message=False, progressbar=False, sleep=0)
        [response] = item.results
        return response

    _FALSE = object()

    def send_blocking(self, cmd, *args, timeout=MINUTE, **kwargs):
        assert timeout is not None
        timeout = timeout or 1

        def _send():
            res = self.send(cmd, *args, **kwargs)
            if res is None:
                return False
            if res is False:
                return self._FALSE
            return res

        res = wait(timeout, _send, sleep=0.01, progressbar=False, throw=False)
        if res == self._FALSE:
            res = False
        return res

    class PipelineExecutor():
        def __init__(self, reactor):
            self.reactor = reactor
            self.commands = []

        def __getattr__(self, item):
            def runner(cmd, *args, **kwargs):
                send = getattr(self.reactor, item)
                self.commands.append(send(cmd, *args, _async=True, **kwargs))

            return runner

        def reset(self):
            self.commands = []

        def execute(self):
            results = [self.reactor.get_response(command) for command in self.commands]
            self.reset()
            return results

    @contextmanager
    def pipeline(self):
        executor = self.PipelineExecutor(self)
        yield executor
        executor.reset()

    def __getattr__(self, cmd):
        if cmd in self.BLOCKING_COMMANDS:
            return partial(self.send_blocking, cmd)
        return partial(self.send, cmd)



import setuptools

with open("README.md") as fh:
    long_description = fh.read()

with open("VERSION") as version_file:
    version = version_file.read().strip()

setuptools.setup(
    name="talker",
    version=version,
    author="Doron Cohen",
    author_email="doron@weka.io",
    description="The almighty remote command executor for Unix machines",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/weka-io/talker",
    license='BSD',
    install_requires=[
        'redis==3.3.7',
        'weka-easypy==0.3.1'
    ],
    extras_require={
        "dev": ["ipython"],
        "test": [
            "fakeredis==1.0.4", "mock==3.0.5",
            # we need to install plumbum in order to make easypy functions -
            # 'initialize_exception_listener' and 'raise_in_main_thread' work properly in client's test
            "plumbum@git+https://github.com/weka-io/plumbum@master#egg=plumbum"
        ],
    },
    packages=["talker", "talker_agent"],
    classifiers=[
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Communications",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "License :: OSI Approved :: BSD License",
    ],
)



import json

from functools import partial

from redis import StrictRedis
import redis.exceptions

from easypy.caching import cached_property, locking_cache
from easypy.collections import chunkify
from easypy.concurrency import MultiObject
from easypy.timing import wait, BackoffTimer, Timer, TimeoutException
from easypy.units import DAY, HOUR, MINUTE
from easypy.humanize import compact

from talker.reactor import TalkerReactor
from talker.command import Cmd, RebootCmd
from talker.errors import RedisTimeoutError, TalkerServerTimeout
from talker.config import (
    REDIS_SOCKET_TIMEOUT, REDIS_SOCKET_CONNECT_TIMEOUT, REDIS_RETRY_ON_TIMEOUT,
    REDIS_HEALTH_CHECK_INTERVAL, AGENT_ACK_TIMEOUT, _logger, _verbose_logger, IS_ALIVE_ACK_TIMEOUT, IS_ALIVE_TIMEOUT
)


@locking_cache
def get_redis(host, password, port):
    return StrictRedis(
        host=host, password=password, port=port,
        socket_timeout=REDIS_SOCKET_TIMEOUT,
        socket_connect_timeout=REDIS_SOCKET_CONNECT_TIMEOUT,
        retry_on_timeout=REDIS_RETRY_ON_TIMEOUT,
        health_check_interval=REDIS_HEALTH_CHECK_INTERVAL
    )


@locking_cache
def get_talker(host, password, port, agent_version=None, name=None):
    return Talker(host, password, port, agent_version=agent_version, name=name)


class Talker(object):
    """
    The client to communicate with Talker agents via the same backend (Redis).

    :param [str] host: The backend host's address
    :param [str] password: The backend user's password
    :param [int] port: The backend port
    :param [str] user: The backend username
    :param [str] agent_version: The Talker agent version we are connecting to
    :param [str] name: Name to use for the Talker client (for logging purposes)
    """

    def __init__(self, host, password, port, agent_version, name):
        self._host = host
        self._password = password
        self._port = port
        self._name = name or "anon"
        self._journal_saved = False
        self.reactor = TalkerReactor(self)
        _logger.debug("%s: initialized", self)

    @property
    def redis_params(self):
        """
        Get the Redis backend's connection parameters

        :returns: The Redis backend's connection parameters
        :rtype: Dict
        """
        return dict(
            host=self._host,
            password=self._password,
            port=int(self._port)
        )

    @cached_property
    def redis(self):
        """
        Get the Redis backend this client is connected to

        :returns: The Redis backend of this Talker client
        :rtype: StrictRedis
        """
        return get_redis(**self.redis_params)

    def __repr__(self):
        return "<{0.__class__.__name__} {0._name}/{0._host}:{0._port}>".format(self)

    def _pipeline_flush_log(self, pipeline):
        _verbose_logger.debug("Sending pipeline: %s", pipeline.command_stack)

    def get_runner_for(self, host_id):
        return partial(self.run, host_id)

    __getitem__ = get_runner_for

    def run(self, host_id, *args,
            name=None,
            ack_timeout=AGENT_ACK_TIMEOUT,
            timeout=HOUR,
            server_timeout=True,
            line_timeout=None,
            log_file=None,
            raise_on_failure=True,
            max_output_per_channel=None, set_new_logpath=None):

        """
        Run a command on a specific host

        :param [str] host_id: The ID of the host to send the command to
        :param [List[str]] args: The command to run on the host (e.g. 'echo', 'hello')
        :param [str] name: A name to give to the command (for logging pruposes)
        :param [int, float] ack_timeout: Agent ack timeout in seconds
        :param [int, float] timeout: Command timeout in seconds
        :param [int, float] server_timeout: Command timeout in seconds to be enforced in the agent
        :param [int, float] line_timeout: Command output timeout in seconds
        :param [str] log_file: Log file path for writing command output on the remote host
        :param [bool] raise_on_failure: Should the client raise an appropriate exception on failure
        :param [int] max_output_per_channel: Maximum command output in bytes
        :param [str] set_new_logpath: Reconfigure the remote agent's log path

        :returns: The command running on the remote host
        :rtype: Cmd
        """

        if not name:
            lines = " ".join(args).strip().splitlines()
            name = compact(lines.pop(0), 60, suffix_length=10)
            if lines:
                name += " ..."
            name = "`%s`" % name

        cmd = Cmd(
            talker=self, host_id=host_id, raise_on_failure=raise_on_failure,
            line_timeout=line_timeout, log_file=log_file,
            args=args, ack_timeout=ack_timeout, timeout=timeout, server_timeout=server_timeout, name=name,
            max_output_per_channel=max_output_per_channel, set_new_logpath=set_new_logpath)
        cmd.send()
        return cmd

    def reboot(self, host_id, timeout=10 * MINUTE, force=RebootCmd.DEFAULT_FORCE, raise_on_failure=True):
        """
        Send a reboot command to the remote host

        :param [str] host_id: The ID of the host to send the command to
        :param [int, float] timeout: Command timeout in seconds
        :param [boot] force: Whether to reboot forcefully (without calling sync)
        :param [bool] raise_on_failure: Should the client raise an appropriate exception on failure

        :returns: The reboot command running on the remote host
        :rtype: RebootCmd
        """
        cmd = RebootCmd(
            force=force,
            talker=self, host_id=host_id, raise_on_failure=raise_on_failure,
            timeout=timeout, server_timeout=False, name="reboot")
        cmd.send()
        return cmd

    def reset_server_error(self, host_id):
        """
        Reset the last error the agent encountered on

        :param [str] host_id: The ID of the host to send the command to
        """
        _logger.debug("Reseting pending error")
        self.reactor.rpush("commands-%s" % host_id, json.dumps(dict(cmd="reset_error")))

    def poll(self, cmds):
        results = {cmd: dict(ack=cmd.ack, retcode=cmd.retcode) for cmd in cmds}

        with self.redis.pipeline() as p:
            res_idx_to_i = {}
            res_idx = 0
            for cmd in cmds:
                if cmd.ack is None:
                    p.lpop(cmd._ack_key)
                    res_idx_to_i[res_idx] = cmd, 'ack'
                    res_idx += 1

                if cmd.retcode is None:
                    p.lpop(cmd._exit_code_key)
                    res_idx_to_i[res_idx] = cmd, 'retcode'
                    res_idx += 1

            if res_idx:
                self._pipeline_flush_log(p)

                with RedisTimeoutError.on_exception(redis.exceptions.TimeoutError, redis=self):
                    pipeline_results = p.execute()

                for (i, result) in enumerate(pipeline_results):
                    cmd, slot = res_idx_to_i[i]
                    results[cmd][slot] = result

        def on_poll(cmd):
            ack = results[cmd]['ack']
            retcode = results[cmd]['retcode']

            if cmd.ack is None and ack:
                cmd.on_ack(ack)

            if retcode is None:
                cmd.check_client_timeout()

            return cmd.on_polled(retcode)

        # Using MultiObject allows to raise multiexception, not just exception on single cmd
        return MultiObject(cmds).call(on_poll)

    def is_done(self, cmds):
        """
        Check if a list of commands are all done

        :param [List[Cmd]] cmds: The list of command to check if are done

        :returns: True if all commands are done else False
        :rtype: bool
        """
        return len(list(filter(lambda i: i is None, self.poll(cmds)))) == 0

    def is_alive(self, host_id, name='is_alive'):
        """
        Do a naive check to see if the agent is responsive.
        If the function returns False it doesn't mean for sure that the agent/machine is down

        :param [str] host_id: The ID of the host to send the command to
        :param [str] name: A name to give to the command (for logging pruposes)
        :return: True if we get ack, False otherwise
        :rtype: bool
        """

        cmd = self.run(host_id, 'true', name=name, timeout=IS_ALIVE_TIMEOUT, ack_timeout=IS_ALIVE_ACK_TIMEOUT)
        try:
            cmd.wait()
        except TalkerServerTimeout:
            pass
        return bool(cmd.ack)

    def wait(self, cmds, sleep=0.5, timeout=DAY, initial_log_interval=12):
        """
        Wait for commands to finish running

        :param [List[Cmd]] cmds: The list of command to wait on
        :param [int, float] sleep: How long to sleep between checks
        :param [int, float] timeout: After how many seconds should we throw a timeout error
        :param [int, float] initial_log_interval: Interval between logging

        :returns: True if all commands are done else False
        :rtype: bool
        """
        for _ in self.as_completed(cmds, sleep, timeout, initial_log_interval):
            pass
        if not all(cmd.retcode in cmd.good_codes for cmd in cmds):  # Not spawning threads without need
            MultiObject(cmds).raise_if_needed()  # as completed omitted all exception, therefore need to re-raise
        return self.poll(cmds)

    def as_completed(self, cmds, sleep=0.5, timeout=DAY, initial_log_interval=12):
        """
        Yields completed commands, until all are completed or the timeout expires (if provided).
        If timeout is 0, this will effectively stop the generator instead of blocking on remaining commands

        :param [List[Cmd]] cmds: The list of command to wait on
        :param [int, float] sleep: How long to sleep between checks
        :param [int, float] timeout: After how many seconds should we throw a timeout error
        :param [int, float] initial_log_interval: Interval between logging

        :returns: Yields commands and their results as a tuple
        :rtype: Tuple[Cmd, result???]
        """
        pending = list(cmds)

        # We prevent exceptions from being raised from this iterator, so that we can keep iterating
        # through all cmds. we'll let the caller deal with raising exceptions mid-iteration if it so desires
        # We also need to deal with the case of re-entering this iterator with incomplete cmds, so we must
        # keep the original 'raise_on_failure' setting across invocations
        for cmd in pending:
            cmd._should_raise_on_failure = getattr(cmd, "_should_raise_on_failure", cmd.raise_on_failure)
            cmd.raise_on_failure = False

        log_interval = initial_log_interval
        logging_timer = BackoffTimer(expiration=log_interval, max_interval=HOUR)
        timeout_timer = Timer(expiration=timeout)

        had_results = False

        def _poll():
            nonlocal had_results
            results = self.poll(pending)

            done_idxes = []
            for i, result in enumerate(results):
                cmd = pending[i]
                if result is not None:
                    done_idxes.append(i)
                    cmd.raise_on_failure = cmd._should_raise_on_failure
                    yield cmd, result

            for i in sorted(done_idxes,
                            reverse=True):  # Alternative would be to copy pending list, but this is more efficient
                pending.pop(i)

            if pending:
                if logging_timer.expired:
                    logging_timer.backoff()
                    MultiObject(pending).check_client_timeout()
                    hosts = sorted(set(cmd.hostname for cmd in pending))
                    _logger.info("Waiting on %s command(s) on %s host(s): %s", len(pending), len(hosts), ", ".join(hosts))
                    for cmd in pending:
                        since_started = (
                            "acked {:ago}".format(cmd.since_started) if cmd.ack else
                            "not-started")
                        _logger.debug("   job-id: %s (%s)", cmd.job_id, since_started)
            if done_idxes:
                had_results = True

        while True:
            yield from _poll()
            if not pending:
                return
            if timeout == 0:
                if had_results:
                    return
                else:
                    continue
            if timeout_timer.expired:
                raise TimeoutException(
                    "Commands (%s) did not complete within timeout" % len(pending),
                    timeout=timeout, pending=pending
                )
            wait(sleep)

    def results(self, cmds, decode='utf-8'):
        """
        Wait for commands to finish and get their results
        This will raise for failed commands if needed

        :param [List[Cmd]] cmds: The list of commands to get their results
        :param [str] decode: Expected output encoding

        :returns: A MultiObject of tuples containing stdout and stderr
        :rtype: MultiObject[Tuple[str, str]]
        """
        self.wait(cmds)
        ret = []
        self.get_output(cmds)
        for i, cmd in enumerate(cmds):
            ret.append(cmd._decode_output(cmd.stdout, decode='utf-8'))
        return ret

    def get_output(self, cmds, decode='utf-8'):
        """
        Get commands' output so far

        :param [List[Cmd]] cmds: The list of commands to get their output
        :param [str] decode: Expected output encoding

        :returns: A MultiObject of tuples containing stdout and stderr
        :rtype: MultiObject[Tuple[str, str]]
        """
        ret = []
        with self.redis.pipeline() as p:
            for cmd in cmds:
                cmd._request_outputs(p)
            self._pipeline_flush_log(p)
            results = p.execute()
            p.reset()
            for i, (stdout, stderr) in enumerate(chunkify(results, 2)):
                cmd = cmds.L[i]
                cmd.on_output(cmd.stdout, stdout)
                cmd.on_output(cmd.stderr, stderr)
                cmd._trim_outputs(stdout, stderr, pipeline=p)
                if decode:
                    (stdout, stderr) = (cmd._decode_output(out, decode) for out in (stdout, stderr))
                ret.append((stdout, stderr))
            self._pipeline_flush_log(p)
            p.execute()
        return MultiObject(ret)

    def iter_output(self, cmds, sleep=1.0, decode='utf-8'):
        """
        Iterate commands output

        :param [List[Cmd]] cmds: The list of commands to iterate on their output
        :param [int, float] sleep: How long to sleep between pollings
        :param [str] decode: Expected output encoding

        :returns: A MultiObject of tuples containing stdout and stderr
        :rtype: MultiObject[Tuple[str, str]]
        """
        while not self.is_done(cmds):
            yield self.get_output(cmds, decode=decode)
            wait(sleep)
        yield self.get_output(cmds, decode=decode)
